# import pdfplumber
import logging
from uuid import uuid5, NAMESPACE_OID
from typing import Optional
from pypdf import PdfReader as pypdf_PdfReader
from cognee.modules.data.chunking import chunk_by_paragraph
from cognee.modules.data.processing.chunk_types.DocumentChunk import DocumentChunk
from .Document import Document

class PdfReader():
    id: str
    file_path: str

    def __init__(self, id: str, file_path: str):
        self.id = id
        self.file_path = file_path

    def get_number_of_pages(self):
        file = pypdf_PdfReader(self.file_path)
        num_pages = file.get_num_pages()
        file.stream.close()
        return num_pages

    def read(self, max_chunk_size: Optional[int] = 1024):
        chunk_index = 0
        chunk_size = 0
        chunked_pages = []
        paragraph_chunks = []

        file = pypdf_PdfReader(self.file_path)

        for (page_index, page) in enumerate(file.pages):
            page_text = page.extract_text()
            chunked_pages.append(page_index)

            for chunk_data in chunk_by_paragraph(page_text, max_chunk_size, batch_paragraphs = True):
                if chunk_size + chunk_data["word_count"] <= max_chunk_size:
                    paragraph_chunks.append(chunk_data)
                    chunk_size += chunk_data["word_count"]
                else:
                    if len(paragraph_chunks) == 0:
                        yield DocumentChunk(
                            text = chunk_data["text"],
                            word_count = chunk_data["word_count"],
                            document_id = str(self.id),
                            chunk_id = str(chunk_data["chunk_id"]),
                            chunk_index = chunk_index,
                            cut_type = chunk_data["cut_type"],
                            pages = [page_index],
                        )
                        paragraph_chunks = []
                        chunk_size = 0
                    else:
                        chunk_text = " ".join(chunk["text"] for chunk in paragraph_chunks)
                        yield DocumentChunk(
                            text = chunk_text,
                            word_count = chunk_size,
                            document_id = str(self.id),
                            chunk_id = str(uuid5(NAMESPACE_OID, f"{str(self.id)}-{chunk_index}")),
                            chunk_index = chunk_index,
                            cut_type = paragraph_chunks[len(paragraph_chunks) - 1]["cut_type"],
                            pages = chunked_pages,
                        )
                        chunked_pages = [page_index]
                        paragraph_chunks = [chunk_data]
                        chunk_size = chunk_data["word_count"]

                    chunk_index += 1

        if len(paragraph_chunks) > 0:
            yield DocumentChunk(
                text = " ".join(chunk["text"] for chunk in paragraph_chunks),
                word_count = chunk_size,
                document_id = str(self.id),
                chunk_id = str(uuid5(NAMESPACE_OID, f"{str(self.id)}-{chunk_index}")),
                chunk_index = chunk_index,
                cut_type = paragraph_chunks[len(paragraph_chunks) - 1]["cut_type"],
                pages = chunked_pages,
            )

        file.stream.close()

class PdfDocument(Document):
    type: str = "pdf"
    title: str
    num_pages: int
    file_path: str

    def __init__(self, title: str, file_path: str):
        self.id = uuid5(NAMESPACE_OID, title)
        self.title = title
        self.file_path = file_path
        logging.debug("file_path: %s", self.file_path)
        reader = PdfReader(self.id, self.file_path)
        self.num_pages = reader.get_number_of_pages()

    def get_reader(self) -> PdfReader:
        logging.debug("file_path: %s", self.file_path)
        reader = PdfReader(self.id, self.file_path)
        return reader

    def to_dict(self) -> dict:
        return dict(
            id = str(self.id),
            type = self.type,
            title = self.title,
            num_pages = self.num_pages,
            file_path = self.file_path,
        )



from uuid import uuid4
from datetime import datetime, timezone
from sqlalchemy import Column, DateTime, UUID, ForeignKey
from cognee.infrastructure.databases.relational import ModelBase

class PipelineTask(ModelBase):
    __tablename__ = "pipeline_task"

    id = Column(UUID, primary_key = True, default = uuid4())

    created_at = Column(DateTime, default = datetime.now(timezone.utc))

    pipeline_id = Column("pipeline", UUID, ForeignKey("pipeline.id"), primary_key = True)
    task_id = Column("task", UUID, ForeignKey("task.id"), primary_key = True)



import dsp
import dspy
from dspy.teleprompt import BootstrapFewShot
from dspy.primitives.example import Example
from cognee.modules.data.extraction.knowledge_graph.extract_knowledge_graph import ExtractKnowledgeGraph
from cognee.root_dir import get_absolute_path
from cognee.infrastructure.files.storage import LocalStorage
from cognee.shared.data_models import Answer
from cognee.infrastructure.llm.get_llm_client import get_llm_client
from cognee.modules.cognify.dataset import HotPotQA
from cognee.infrastructure.llm import get_llm_config

def train():
    colbertv2_wiki17_abstracts = dspy.ColBERTv2(url = "http://20.102.90.50:2017/wiki17_abstracts")

    dspy.configure(rm = colbertv2_wiki17_abstracts)

    def evaluate_answer(example, graph_prediction, trace = None):
        llm_client = get_llm_client()

        try:
            answer_prediction = llm_client.create_structured_output(
                text_input = example.question,
                system_prompt = f"""Answer the question by looking at the provided knowledge graph.
                Use only the graph to answer the question and be very brief.
                This is the knowledge graph:
                {graph_prediction.graph.model_dump(mode = "json")}""",
                response_model = Answer,
            )
        except:
            return False

        return dsp.answer_match(example.answer, [answer_prediction.answer], frac = 0.8) or \
            dsp.passage_match([example.answer], [answer_prediction.answer])

    optimizer = BootstrapFewShot(metric = evaluate_answer)

    dataset = HotPotQA(
        train_seed = 1,
        train_size = 16,
        eval_seed = 2023,
        dev_size = 8,
        test_size = 0,
        keep_details = True,
    )

    # Train
    train_examples = [
        Example(
            base = None,
            question = example.question,
            context = "\r\n".join("".join(sentences) for sentences in example.context["sentences"]),
            answer = example.answer,
        ) for example in dataset.train
    ]

    trainset = [example.with_inputs("context", "question") for example in train_examples]

    llm_config = get_llm_config()
    gpt4 = dspy.OpenAI(model = llm_config.llm_model, api_key = llm_config.llm_api_key, model_type = "chat", max_tokens = 4096)

    compiled_extract_knowledge_graph = optimizer.compile(ExtractKnowledgeGraph(lm = gpt4), trainset = trainset)

    # Save program
    LocalStorage.ensure_directory_exists(get_absolute_path("./programs/extract_knowledge_graph"))
    compiled_extract_knowledge_graph.save(get_absolute_path("./programs/extract_knowledge_graph/extract_knowledge_graph.json"))

if __name__ == "__main__":
    train()



from .extract_content_graph import extract_content_graph

async def extract_knowledge_graph(text: str, cognitive_layer, graph_model):
    return await extract_content_graph(text, cognitive_layer, graph_model)




from typing import TypedDict
from pydantic import BaseModel, Field
from cognee.infrastructure.databases.vector import DataPoint, get_vector_engine

class TextChunk(TypedDict):
    text: str
    chunk_id: str
    file_metadata: dict

async def add_data_chunks(dataset_data_chunks: dict[str, list[TextChunk]]):
    vector_engine = get_vector_engine()

    identified_chunks = []

    class PayloadSchema(BaseModel):
        text: str = Field(...)

    for (dataset_name, chunks) in dataset_data_chunks.items():
        try:

            await vector_engine.create_collection(dataset_name, payload_schema = PayloadSchema)
        except Exception as error:
            print(error)
            pass

        dataset_chunks = [
            dict(
                chunk_id = chunk["chunk_id"],
                collection = dataset_name,
                text = chunk["text"],
                document_id = chunk["document_id"],
                file_metadata = chunk["file_metadata"],
            ) for chunk in chunks
        ]

        identified_chunks.extend(dataset_chunks)

        await vector_engine.create_data_points(
            dataset_name,
            [
                DataPoint[PayloadSchema](
                    id = chunk["chunk_id"],
                    payload = PayloadSchema.parse_obj(dict(text = chunk["text"])),
                    embed_field = "text",
                ) for chunk in dataset_chunks
            ],
        )

    return identified_chunks


async def add_data_chunks_basic_rag(dataset_data_chunks: dict[str, list[TextChunk]]):
    vector_engine = get_vector_engine()

    identified_chunks = []

    class PayloadSchema(BaseModel):
        text: str = Field(...)

    for (dataset_name, chunks) in dataset_data_chunks.items():
        try:

            await vector_engine.create_collection("basic_rag", payload_schema = PayloadSchema)
        except Exception as error:
            print(error)

        dataset_chunks = [
            dict(
                chunk_id = chunk["chunk_id"],
                collection = "basic_rag",
                text = chunk["text"],
                document_id = chunk["document_id"],
                file_metadata = chunk["file_metadata"],
            ) for chunk in chunks
        ]

        identified_chunks.extend(dataset_chunks)

        await vector_engine.create_data_points(
            "basic_rag",
            [
                DataPoint[PayloadSchema](
                    id = chunk["chunk_id"],
                    payload = PayloadSchema.parse_obj(dict(text = chunk["text"])),
                    embed_field = "text",
                ) for chunk in dataset_chunks
            ],
        )

    return identified_chunks



from typing import Type
from pydantic import BaseModel
from cognee.shared.data_models import KnowledgeGraph
from cognee.infrastructure.databases.graph import get_graph_engine
from ...processing.chunk_types.DocumentChunk import DocumentChunk
from .add_model_class_to_graph import add_model_class_to_graph

async def establish_graph_topology(data_chunks: list[DocumentChunk], topology_model: Type[BaseModel]):
    if topology_model == KnowledgeGraph:
        return data_chunks

    graph_engine = await get_graph_engine()

    await add_model_class_to_graph(topology_model, graph_engine)

    return data_chunks


def generate_node_id(node_id: str) -> str:
    return node_id.upper().replace(" ", "_").replace("'", "")



from cognee.infrastructure.databases.relational.config import get_relationaldb_config

def update_task_status(data_id: str, status: str):
    config = get_relationaldb_config()
    db_engine = config.database_engine
    db_engine.insert_data("cognee.cognee", "cognee_task_status", [dict(data_id = data_id, status = status)])



from typing import Union, Dict
import networkx as nx
from cognee.infrastructure.databases.graph import get_graph_engine
from cognee.infrastructure.databases.vector import get_vector_engine

async def search_adjacent(graph: Union[nx.Graph, any], query: str, other_param: dict = None) -> Dict[str, str]:
    """
    Find the neighbours of a given node in the graph and return their ids and descriptions.

    Parameters:
    - graph (Union[nx.Graph, AsyncSession]): The graph object or Neo4j session.
    - query (str): Unused in this implementation but could be used for future enhancements.
    - other_param (dict, optional): A dictionary that may contain 'node_id' to specify the node.

    Returns:
    - Dict[str, str]: A dictionary containing the unique identifiers and descriptions of the neighbours of the given node.
    """
    node_id = other_param.get("node_id") if other_param else query

    if node_id is None:
        return {}

    graph_engine = await get_graph_engine()

    exact_node = await graph_engine.extract_node(node_id)
    if exact_node is not None and "id" in exact_node:
        neighbours = await graph_engine.get_neighbours(exact_node["id"])
    else:
        vector_engine = get_vector_engine()
        collection_name = "classification"
        data_points = await vector_engine.search(collection_name, query_text = node_id, limit = 5)

        if len(data_points) == 0:
            return []

        neighbours = await graph_engine.get_neighbours(data_points[0].id)

    return [node["name"] for node in neighbours]



import asyncio
from cognee.modules.pipelines.operations.run_tasks import run_tasks
from cognee.modules.pipelines.tasks.Task import Task


async def main():
    def number_generator(num):
        for i in range(num):
            yield i + 1

    async def add_one(num):
        yield num + 1

    async def multiply_by_two(nums):
        for num in nums:
            yield num * 2

    async def add_one_to_batched_data(num):
        yield num + 1

    pipeline = run_tasks([
        Task(number_generator, task_config = {"batch_size": 1}),
        Task(add_one, task_config = {"batch_size": 5}),
        Task(multiply_by_two, task_config = {"batch_size": 1}),
        Task(add_one_to_batched_data),
    ], 10)

    async for result in pipeline:
        print("\n")
        print(result)
        print("\n")

if __name__ == "__main__":
    asyncio.run(main())



from typing import List
from uuid import uuid4
from datetime import datetime, timezone
from sqlalchemy import Column, UUID, DateTime, String, Text
from sqlalchemy.orm import relationship, Mapped
from cognee.infrastructure.databases.relational import ModelBase
from .PipelineTask import PipelineTask

class Pipeline(ModelBase):
    __tablename__ = "pipelines"

    id = Column(UUID, primary_key = True, default = uuid4())
    name = Column(String)
    description = Column(Text, nullable = True)

    created_at = Column(DateTime, default = datetime.now(timezone.utc))
    updated_at = Column(DateTime, onupdate = datetime.now(timezone.utc))

    tasks = Mapped[List["Task"]] = relationship(
        secondary = PipelineTask.__tablename__,
        back_populates = "pipeline",
    )



from cognee.infrastructure.databases.graph.get_graph_engine import get_graph_engine
from cognee.infrastructure.databases.vector import get_vector_engine

async def search_traverse(query: str, graph): # graph must be there in order to be compatible with generic call
    graph_engine = await get_graph_engine()
    vector_engine = get_vector_engine()

    results = await vector_engine.search("classification", query_text = query, limit = 10)

    rules = []

    if len(results) > 0:
        for result in results:
            graph_node_id = result.id

            edges = await graph_engine.get_edges(graph_node_id)

            for edge in edges:
                rules.append(f"{edge[0]} {edge[2]['relationship_name']} {edge[1]}")

    return rules



from uuid import uuid5, NAMESPACE_OID
from typing import Optional, Generator

from cognee.infrastructure.llm.get_llm_client import get_llm_client
from cognee.modules.data.chunking import chunk_by_paragraph
from cognee.modules.data.processing.chunk_types.DocumentChunk import DocumentChunk
from cognee.modules.data.processing.document_types.Document import Document

class AudioReader:
    id: str
    file_path: str

    def __init__(self, id: str, file_path: str):
        self.id = id
        self.file_path = file_path
        self.llm_client = get_llm_client()  # You can choose different models like "tiny", "base", "small", etc.


    def read(self, max_chunk_size: Optional[int] = 1024):
        chunk_index = 0
        chunk_size = 0
        chunked_pages = []
        paragraph_chunks = []

        # Transcribe the audio file
        result = self.llm_client.create_transcript(self.file_path)
        text = result.text

        # Simulate reading text in chunks as done in TextReader
        def read_text_chunks(text, chunk_size):
            for i in range(0, len(text), chunk_size):
                yield text[i:i + chunk_size]

        page_index = 0

        for page_text in read_text_chunks(text, max_chunk_size):
            chunked_pages.append(page_index)
            page_index += 1

            for chunk_data in chunk_by_paragraph(page_text, max_chunk_size, batch_paragraphs=True):
                if chunk_size + chunk_data["word_count"] <= max_chunk_size:
                    paragraph_chunks.append(chunk_data)
                    chunk_size += chunk_data["word_count"]
                else:
                    if len(paragraph_chunks) == 0:
                        yield DocumentChunk(
                            text=chunk_data["text"],
                            word_count=chunk_data["word_count"],
                            document_id=str(self.id),
                            chunk_id=str(chunk_data["chunk_id"]),
                            chunk_index=chunk_index,
                            cut_type=chunk_data["cut_type"],
                            pages=[page_index],
                        )
                        paragraph_chunks = []
                        chunk_size = 0
                    else:
                        chunk_text = " ".join(chunk["text"] for chunk in paragraph_chunks)
                        yield DocumentChunk(
                            text=chunk_text,
                            word_count=chunk_size,
                            document_id=str(self.id),
                            chunk_id=str(uuid5(NAMESPACE_OID, f"{str(self.id)}-{chunk_index}")),
                            chunk_index=chunk_index,
                            cut_type=paragraph_chunks[len(paragraph_chunks) - 1]["cut_type"],
                            pages=chunked_pages,
                        )
                        chunked_pages = [page_index]
                        paragraph_chunks = [chunk_data]
                        chunk_size = chunk_data["word_count"]

                    chunk_index += 1

        if len(paragraph_chunks) > 0:
            yield DocumentChunk(
                text=" ".join(chunk["text"] for chunk in paragraph_chunks),
                word_count=chunk_size,
                document_id=str(self.id),
                chunk_id=str(uuid5(NAMESPACE_OID, f"{str(self.id)}-{chunk_index}")),
                chunk_index=chunk_index,
                cut_type=paragraph_chunks[len(paragraph_chunks) - 1]["cut_type"],
                pages=chunked_pages,
            )

class AudioDocument(Document):
    type: str = "audio"
    title: str
    file_path: str

    def __init__(self, title: str, file_path: str):
        self.id = uuid5(NAMESPACE_OID, title)
        self.title = title
        self.file_path = file_path

        reader = AudioReader(self.id, self.file_path)

    def get_reader(self) -> AudioReader:
        reader = AudioReader(self.id, self.file_path)
        return reader

    def to_dict(self) -> dict:
        return dict(
            id=str(self.id),
            type=self.type,
            title=self.title,
            file_path=self.file_path,
        )


# if __name__ == "__main__":
#     # Sample usage of AudioDocument
#     audio_document = AudioDocument("sample_audio", "/Users/vasa/Projects/cognee/cognee/modules/data/processing/document_types/preamble10.wav")
#     audio_reader = audio_document.get_reader()
#     for chunk in audio_reader.read():
#         print(chunk.text)
#         print(chunk.word_count)
#         print(chunk.document_id)
#         print(chunk.chunk_id)
#         print(chunk.chunk_index)
#         print(chunk.cut_type)
#         print(chunk.pages)
#         print("----")



from cognee.infrastructure.files.storage import LocalStorage
from cognee.infrastructure.databases.relational import DuckDBAdapter

def create_relational_engine(db_path: str, db_name: str):
    LocalStorage.ensure_directory_exists(db_path)

    return DuckDBAdapter(
        db_name = db_name,
        db_path = db_path,
    )



import logging
from cognee.modules.topology.extraction.extract_topology import extract_topology
from cognee.infrastructure.databases.graph.config import get_graph_config

logger = logging.getLogger(__name__)

async def infer_data_topology(content: str, graph_topology=None):
    if graph_topology is None:
        graph_config = get_graph_config()
        graph_topology = graph_config.graph_topology

    print("content: ", type(content))
    try:
        return (await extract_topology(
            content,
            graph_topology
        ))
    except Exception as error:
        logger.error("Error extracting topology from content: %s", error, exc_info = True)
        raise error



from cognee.infrastructure.databases.relational.config import get_relationaldb_config

def get_task_status(data_ids: [str]):
    relational_config = get_relationaldb_config()
    db_engine = relational_config.database_engine

    formatted_data_ids = ", ".join([f"'{data_id}'" for data_id in data_ids])

    datasets_statuses = db_engine.execute_query(
      f"""SELECT data_id, status
      FROM (
          SELECT data_id, status, ROW_NUMBER() OVER (PARTITION BY data_id ORDER BY created_at DESC) as rn
          FROM cognee.cognee.cognee_task_status
          WHERE data_id IN ({formatted_data_ids})
      ) t
      WHERE rn = 1;"""
    )

    return { dataset["data_id"]: dataset["status"] for dataset in datasets_statuses }




import networkx as nx
from typing import Union
from cognee.shared.data_models import GraphDBType
from cognee.infrastructure.databases.graph.config import get_graph_config

async def search_cypher(query:str, graph: Union[nx.Graph, any]):
    """
    Use a Cypher query to search the graph and return the results.
    """
    graph_config = get_graph_config()

    if graph_config.graph_database_provider == "neo4j":
        result = await graph.run(query)
        return result

    else:
        raise ValueError("Unsupported graph engine type.")



from cognee.infrastructure.databases.relational.config import get_relationaldb_config

def create_task_status_table():
    config = get_relationaldb_config()
    db_engine = config.database_engine

    db_engine.create_table("cognee.cognee", "cognee_task_status", [
        dict(name = "data_id", type = "STRING"),
        dict(name = "status", type = "STRING"),
        dict(name = "created_at", type = "TIMESTAMP DEFAULT CURRENT_TIMESTAMP"),
    ])



from uuid import UUID
from cognee.shared.data_models import Document
from cognee.infrastructure.databases.graph import get_graph_engine

async def save_document_node(document: Document, parent_node_id: UUID = None):
    graph_engine = get_graph_engine()

    await graph_engine.add_node(document.id, document.model_dump())

    if parent_node_id:
        await graph_engine.add_edge(
            parent_node_id,
            document.id,
            "has_document",
            dict(relationship_name = "has_document"),
        )

    return document



from .ModelBase import ModelBase
from .DatabaseEngine import DatabaseEngine
from .sqlite.SqliteEngine import SqliteEngine
from .duckdb.DuckDBAdapter import DuckDBAdapter
from .config import get_relationaldb_config



from typing import Union, Dict
import networkx as nx
from cognee.infrastructure.databases.vector import get_vector_engine

async def search_similarity(query: str, graph: Union[nx.Graph, any]) -> Dict[str, str]:
    """
    Parameters:
    - graph (Union[nx.Graph, AsyncSession]): The graph object or Neo4j session.
    - query (str): The query string to filter nodes by, e.g., 'SUMMARY'.

    Returns:
    - Dict[str, str]: A dictionary where keys are node identifiers containing the query string, and values are their 'result' attributes.
    """
    vector_engine = get_vector_engine()

    similar_results = await vector_engine.search("chunks", query, limit = 5)
    results = [{
        "text": result.payload["text"],
        "chunk_id": result.payload["chunk_id"],
      } for result in similar_results]

    return results



import asyncio
import logging
from typing import Union

from cognee.infrastructure.databases.graph import get_graph_config
from cognee.modules.cognify.config import get_cognify_config
from cognee.infrastructure.databases.relational.config import get_relationaldb_config
from cognee.modules.data.processing.document_types.AudioDocument import AudioDocument
from cognee.modules.data.processing.document_types.ImageDocument import ImageDocument
from cognee.shared.data_models import KnowledgeGraph
from cognee.modules.data.processing.document_types import PdfDocument, TextDocument
from cognee.modules.cognify.vector import save_data_chunks
from cognee.modules.data.processing.process_documents import process_documents
from cognee.modules.classification.classify_text_chunks import classify_text_chunks
from cognee.modules.data.extraction.data_summary.summarize_text_chunks import summarize_text_chunks
from cognee.modules.data.processing.filter_affected_chunks import filter_affected_chunks
from cognee.modules.data.processing.remove_obsolete_chunks import remove_obsolete_chunks
from cognee.modules.data.extraction.knowledge_graph.expand_knowledge_graph import expand_knowledge_graph
from cognee.modules.data.extraction.knowledge_graph.establish_graph_topology import establish_graph_topology
from cognee.modules.pipelines.tasks.Task import Task
from cognee.modules.pipelines import run_tasks, run_tasks_parallel
from cognee.modules.tasks import create_task_status_table, update_task_status, get_task_status

logger = logging.getLogger("cognify.v2")

update_status_lock = asyncio.Lock()

async def cognify(datasets: Union[str, list[str]] = None, root_node_id: str = None):
    relational_config = get_relationaldb_config()
    db_engine = relational_config.database_engine
    create_task_status_table()

    if datasets is None or len(datasets) == 0:
        return await cognify(db_engine.get_datasets())


    async def run_cognify_pipeline(dataset_name: str, files: list[dict]):
        async with update_status_lock:
            task_status = get_task_status([dataset_name])

            if dataset_name in task_status and task_status[dataset_name] == "DATASET_PROCESSING_STARTED":
                logger.info(f"Dataset {dataset_name} is being processed.")
                return

            update_task_status(dataset_name, "DATASET_PROCESSING_STARTED")
        try:
            cognee_config = get_cognify_config()
            graph_config = get_graph_config()
            root_node_id = None

            if graph_config.infer_graph_topology and graph_config.graph_topology_task:
                from cognee.modules.topology.topology import TopologyEngine
                topology_engine = TopologyEngine(infer=graph_config.infer_graph_topology)
                root_node_id = await topology_engine.add_graph_topology(files = files)
            elif graph_config.infer_graph_topology and not graph_config.infer_graph_topology:
                from cognee.modules.topology.topology import TopologyEngine
                topology_engine = TopologyEngine(infer=graph_config.infer_graph_topology)
                await topology_engine.add_graph_topology(graph_config.topology_file_path)
            elif not graph_config.graph_topology_task:
                root_node_id = "ROOT"

            tasks = [
                Task(process_documents, parent_node_id = root_node_id, task_config = { "batch_size": 10 }), # Classify documents and save them as a nodes in graph db, extract text chunks based on the document type
                Task(establish_graph_topology, topology_model = KnowledgeGraph), # Set the graph topology for the document chunk data
                Task(expand_knowledge_graph, graph_model = KnowledgeGraph), # Generate knowledge graphs from the document chunks and attach it to chunk nodes
                Task(filter_affected_chunks, collection_name = "chunks"), # Find all affected chunks, so we don't process unchanged chunks
                Task(
                    save_data_chunks,
                    collection_name = "chunks",
                ), # Save the document chunks in vector db and as nodes in graph db (connected to the document node and between each other)
                run_tasks_parallel([
                    Task(
                        summarize_text_chunks,
                        summarization_model = cognee_config.summarization_model,
                        collection_name = "chunk_summaries",
                    ), # Summarize the document chunks
                    Task(
                        classify_text_chunks,
                        classification_model = cognee_config.classification_model,
                    ),
                ]),
                Task(remove_obsolete_chunks), # Remove the obsolete document chunks.
            ]

            pipeline = run_tasks(tasks, [
                PdfDocument(title=f"{file['name']}.{file['extension']}", file_path=file["file_path"]) if file["extension"] == "pdf" else
                AudioDocument(title=f"{file['name']}.{file['extension']}", file_path=file["file_path"]) if file["extension"] == "audio" else
                ImageDocument(title=f"{file['name']}.{file['extension']}", file_path=file["file_path"]) if file["extension"] == "image" else
                TextDocument(title=f"{file['name']}.{file['extension']}", file_path=file["file_path"])
                for file in files
            ])

            async for result in pipeline:
                print(result)

            update_task_status(dataset_name, "DATASET_PROCESSING_FINISHED")
        except Exception as error:
            update_task_status(dataset_name, "DATASET_PROCESSING_ERROR")
            raise error


    existing_datasets = db_engine.get_datasets()

    awaitables = []


    # dataset_files = []
    # dataset_name = datasets.replace(".", "_").replace(" ", "_")

    # for added_dataset in existing_datasets:
    #     if dataset_name in added_dataset:
    #         dataset_files.append((added_dataset, db_engine.get_files_metadata(added_dataset)))

    for dataset in datasets:
        if dataset in existing_datasets:
            # for file_metadata in files:
            #     if root_node_id is None:
            #         root_node_id=file_metadata['id']
            awaitables.append(run_cognify_pipeline(dataset, db_engine.get_files_metadata(dataset)))

    return await asyncio.gather(*awaitables)


#
# if __name__ == "__main__":
#     from cognee.api.v1.add import add
#     from cognee.api.v1.datasets.datasets import datasets
#
#
#     async def aa():
#         await add("TEXT ABOUT NLP AND MONKEYS")
#
#         print(datasets.discover_datasets())
#
#         return



    # asyncio.run(cognify())



import logging
from .extraction.extract_categories import extract_categories
from cognee.modules.cognify.config import get_cognify_config

config = get_cognify_config()
logger = logging.getLogger(__name__)

async def get_content_categories(content: str):
    try:
        return await extract_categories(
            content,
            config.classification_model
        )
    except Exception as error:
        logger.error("Error extracting categories from content: %s", error, exc_info = True)
        raise error



import asyncio
from uuid import uuid4
from typing import List, Union
import logging
import nltk
from asyncio import Lock
from nltk.corpus import stopwords

from cognee.infrastructure.data.chunking.get_chunking_engine import get_chunk_engine
from cognee.infrastructure.databases.graph.config import get_graph_config
from cognee.infrastructure.databases.vector.embeddings.LiteLLMEmbeddingEngine import LiteLLMEmbeddingEngine
from cognee.modules.cognify.graph.add_node_connections import group_nodes_by_layer, \
    graph_ready_output, connect_nodes_in_graph
from cognee.modules.cognify.graph.add_data_chunks import add_data_chunks
from cognee.modules.cognify.graph.add_document_node import add_document_node
from cognee.modules.cognify.graph.add_classification_nodes import add_classification_nodes
from cognee.modules.cognify.graph.add_cognitive_layer_graphs import add_cognitive_layer_graphs
from cognee.modules.cognify.graph.add_summary_nodes import add_summary_nodes
from cognee.modules.cognify.llm.resolve_cross_graph_references import resolve_cross_graph_references
from cognee.infrastructure.databases.graph import get_graph_engine
from cognee.modules.cognify.graph.add_cognitive_layers import add_cognitive_layers
from cognee.infrastructure.files.utils.guess_file_type import guess_file_type, FileTypeException
from cognee.infrastructure.files.utils.extract_text_from_file import extract_text_from_file
from cognee.modules.data.get_content_categories import get_content_categories
from cognee.modules.data.get_content_summary import get_content_summary
from cognee.modules.data.get_cognitive_layers import get_cognitive_layers
from cognee.modules.data.get_layer_graphs import get_layer_graphs
from cognee.shared.data_models import KnowledgeGraph, ChunkStrategy, ChunkEngine
from cognee.shared.utils import send_telemetry
from cognee.modules.tasks import create_task_status_table, update_task_status
from cognee.shared.SourceCodeGraph import SourceCodeGraph
from cognee.modules.tasks import get_task_status
from cognee.infrastructure.data.chunking.config import get_chunk_config
from cognee.modules.cognify.config import get_cognify_config
from cognee.infrastructure.databases.relational.config import get_relationaldb_config

USER_ID = "default_user"

logger = logging.getLogger("cognify")

update_status_lock = Lock()

async def cognify(datasets: Union[str, List[str]] = None):
    """This function is responsible for the cognitive processing of the content."""
    # Has to be loaded in advance, multithreading doesn't work without it.
    nltk.download("stopwords", quiet=True)
    stopwords.ensure_loaded()
    create_task_status_table()

    graph_client = await get_graph_engine()

    relational_config = get_relationaldb_config()
    db_engine = relational_config.database_engine

    if datasets is None or len(datasets) == 0:
        datasets = db_engine.get_datasets()

    awaitables = []

    async def handle_cognify_task(dataset_name: str):
        async with update_status_lock:
            task_status = get_task_status([dataset_name])

            if dataset_name in task_status and task_status[dataset_name] == "DATASET_PROCESSING_STARTED":
                logger.info(f"Dataset {dataset_name} is being processed.")
                return

            update_task_status(dataset_name, "DATASET_PROCESSING_STARTED")

        try:
            await cognify(dataset_name)
            update_task_status(dataset_name, "DATASET_PROCESSING_FINISHED")
        except Exception as error:
            update_task_status(dataset_name, "DATASET_PROCESSING_ERROR")
            raise error


    # datasets is a list of dataset names
    if isinstance(datasets, list):
        for dataset_name in datasets:
            awaitables.append(handle_cognify_task(dataset_name))

        graphs = await asyncio.gather(*awaitables)
        return graphs[0]

    added_datasets = db_engine.get_datasets()

    # datasets is a dataset name string
    dataset_files = []
    dataset_name = datasets.replace(".", "_").replace(" ", "_")

    for added_dataset in added_datasets:
        if dataset_name in added_dataset:
            dataset_files.append((added_dataset, db_engine.get_files_metadata(added_dataset)))

    chunk_config = get_chunk_config()
    chunk_engine = get_chunk_engine()
    chunk_strategy = chunk_config.chunk_strategy

    async def process_batch(files_batch):
        data_chunks = {}

        for dataset_name, file_metadata, document_id in files_batch:
            with open(file_metadata["file_path"], "rb") as file:
                try:
                    file_type = guess_file_type(file)
                    text = extract_text_from_file(file, file_type)
                    if text is None:
                        text = "empty file"
                    if text == "":
                        text = "empty file"
                    subchunks,_ = chunk_engine.chunk_data(chunk_strategy, text, chunk_config.chunk_size, chunk_config.chunk_overlap)

                    if dataset_name not in data_chunks:
                        data_chunks[dataset_name] = []

                    for subchunk in subchunks:
                        data_chunks[dataset_name].append(dict(
                            document_id = document_id,
                            chunk_id = str(uuid4()),
                            text = subchunk,
                            file_metadata = file_metadata,
                        ))

                except FileTypeException:
                    logger.warning("File (%s) has an unknown file type. We are skipping it.", file_metadata["id"])

        added_chunks = await add_data_chunks(data_chunks)
        # await add_data_chunks_basic_rag(data_chunks)

        await asyncio.gather(
            *[process_text(
              chunk["collection"],
              chunk["chunk_id"],
              chunk["text"],
              chunk["file_metadata"],
              chunk["document_id"]
            ) for chunk in added_chunks]
        )

    batch_size = 20
    file_count = 0
    files_batch = []

    graph_config = get_graph_config()

    if graph_config.infer_graph_topology and graph_config.graph_topology_task:
        from cognee.modules.topology.topology import TopologyEngine
        topology_engine = TopologyEngine(infer=graph_config.infer_graph_topology)
        await topology_engine.add_graph_topology(dataset_files=dataset_files)
    elif not graph_config.infer_graph_topology:
        from cognee.modules.topology.topology import TopologyEngine
        topology_engine = TopologyEngine(infer=graph_config.infer_graph_topology)
        await topology_engine.add_graph_topology(graph_config.topology_file_path)
    elif not graph_config.graph_topology_task:
        parent_node_id = f"DefaultGraphModel__{USER_ID}"


    for (dataset_name, files) in dataset_files:
        for file_metadata in files:
            if parent_node_id:
                document_id = await add_document_node(
                    graph_client,
                    parent_node_id = parent_node_id,
                    document_metadata = file_metadata,
                )
            else:
                document_id = await add_document_node(
                    graph_client,
                    parent_node_id = file_metadata['id'],
                    document_metadata = file_metadata,
                )

            files_batch.append((dataset_name, file_metadata, document_id))
            file_count += 1

            if file_count >= batch_size:
                await process_batch(files_batch)
                files_batch = []
                file_count = 0

    # Process any remaining files in the last batch
    if len(files_batch) > 0:
        await process_batch(files_batch)

    return graph_client.graph


async def process_text(chunk_collection: str, chunk_id: str, input_text: str, file_metadata: dict, document_id: str):
    print(f"Processing chunk ({chunk_id}) from document ({file_metadata['id']}).")

    graph_config = get_graph_config()
    graph_client = await get_graph_engine()
    graph_topology = graph_config.graph_model

    if graph_topology == SourceCodeGraph:
        classified_categories = [{"data_type": "text", "category_name": "Code and functions"}]
    elif graph_topology == KnowledgeGraph:
        classified_categories = await get_content_categories(input_text)
    else:
        classified_categories = [{"data_type": "text", "category_name": "Unclassified text"}]

    await add_classification_nodes(
        graph_client,
        parent_node_id = document_id,
        categories = classified_categories,
    )
    print(f"Chunk ({chunk_id}) classified.")

    content_summary = await get_content_summary(input_text)
    await add_summary_nodes(graph_client, document_id, content_summary)
    print(f"Chunk ({chunk_id}) summarized.")

    cognify_config = get_cognify_config()

    cognitive_layers = await get_cognitive_layers(input_text, classified_categories)
    cognitive_layers = cognitive_layers[:cognify_config.cognitive_layers_limit]

    try:
        cognitive_layers = (await add_cognitive_layers(graph_client, document_id, cognitive_layers))[:2]
        print("cognitive_layers", cognitive_layers)
        layer_graphs = await get_layer_graphs(input_text, cognitive_layers)
        await add_cognitive_layer_graphs(graph_client, chunk_collection, chunk_id, layer_graphs)
    except:
        pass


    if cognify_config.connect_documents is True:
        db_engine = get_relationaldb_config().database_engine
        relevant_documents_to_connect = db_engine.fetch_cognify_data(excluded_document_id = document_id)

        list_of_nodes = []

        relevant_documents_to_connect.append({
            "layer_id": document_id,
        })

        for document in relevant_documents_to_connect:
            node_descriptions_to_match = await graph_client.extract_node_description(document["layer_id"])
            list_of_nodes.extend(node_descriptions_to_match)

        nodes_by_layer = await group_nodes_by_layer(list_of_nodes)

        results = await resolve_cross_graph_references(nodes_by_layer)

        relationships = graph_ready_output(results)

        await connect_nodes_in_graph(
            graph_client,
            relationships,
            score_threshold = cognify_config.intra_layer_score_treshold
        )

    send_telemetry("cognee.cognify")

    print(f"Chunk ({chunk_id}) cognified.")



if __name__ == "__main__":

    async def test():
        # await prune.prune_system()
        # #
        # from cognee.api.v1.add import add
        # data_directory_path = os.path.abspath("../../.data")
        # # print(data_directory_path)
        # # config.data_root_directory(data_directory_path)
        # # cognee_directory_path = os.path.abspath(".cognee_system")
        # # config.system_root_directory(cognee_directory_path)
        #
        # await add("data://" +data_directory_path, "example")

        text = """Conservative PP in the lead in Spain, according to estimate
                An estimate has been published for Spain:
                
                Opposition leader Alberto Núñez Feijóo’s conservative People’s party (PP): 32.4%
                
                Spanish prime minister Pedro Sánchez’s Socialist party (PSOE): 30.2%
                
                The far-right Vox party: 10.4%
                
                In Spain, the right has sought to turn the European election into a referendum on Sánchez.
                
                Ahead of the vote, public attention has focused on a saga embroiling the prime minister’s wife, Begoña Gómez, who is being investigated over allegations of corruption and influence-peddling, which Sanchez has dismissed as politically-motivated and totally baseless."""

        from cognee.api.v1.add import add

        await add([text], "example_dataset")

        from cognee.api.v1.config.config import config
        config.set_chunk_engine(ChunkEngine.LANGCHAIN_ENGINE )
        config.set_chunk_strategy(ChunkStrategy.LANGCHAIN_CHARACTER)
        config.embedding_engine = LiteLLMEmbeddingEngine()

        graph = await cognify()
        # vector_client = infrastructure_config.get_config("vector_engine")
        #
        # out = await vector_client.search(collection_name ="basic_rag", query_text="show_all_processes", limit=10)
        #
        # print("results", out)
        #
        # from cognee.shared.utils import render_graph
        #
        # await render_graph(graph, include_color=True, include_nodes=False, include_size=False)

    import asyncio
    asyncio.run(test())



from duckdb import CatalogException
from cognee.modules.ingestion import discover_directory_datasets
from cognee.modules.tasks import get_task_status
from cognee.infrastructure.databases.relational.config import get_relationaldb_config

class datasets():
    @staticmethod
    def list_datasets():
        relational_config = get_relationaldb_config()
        db = relational_config.database_engine
        return db.get_datasets()

    @staticmethod
    def discover_datasets(directory_path: str):
        return list(discover_directory_datasets(directory_path).keys())

    @staticmethod
    def list_data(dataset_name: str):
        relational_config = get_relationaldb_config()
        db = relational_config.database_engine
        try:
            return db.get_files_metadata(dataset_name)
        except CatalogException:
            return None

    @staticmethod
    def get_status(dataset_ids: list[str]) -> dict:
        try:
            return get_task_status(dataset_ids)
        except CatalogException:
            return {}

    @staticmethod
    def delete_dataset(dataset_id: str):
        relational_config = get_relationaldb_config()
        db = relational_config.database_engine
        try:
            return db.delete_table(dataset_id)
        except CatalogException:
            return {}



import asyncio
from datetime import datetime
from typing import Type
from pydantic import BaseModel
from cognee.infrastructure.databases.graph import get_graph_engine
from ...processing.chunk_types.DocumentChunk import DocumentChunk
from .extract_knowledge_graph import extract_content_graph

async def expand_knowledge_graph(data_chunks: list[DocumentChunk], graph_model: Type[BaseModel]):
    chunk_graphs = await asyncio.gather(
        *[extract_content_graph(chunk.text, graph_model) for chunk in data_chunks]
    )

    graph_engine = await get_graph_engine()

    type_ids = [generate_node_id(node.type) for chunk_graph in chunk_graphs for node in chunk_graph.nodes]
    graph_type_node_ids = list(set(type_ids))
    graph_type_nodes = await graph_engine.extract_nodes(graph_type_node_ids)
    existing_type_nodes_map = {node["id"]: node for node in graph_type_nodes}

    graph_nodes = []
    graph_edges = []

    for (chunk_index, chunk) in enumerate(data_chunks):
        graph = chunk_graphs[chunk_index]
        if graph is None:
            continue

        for node in graph.nodes:
            node_id = generate_node_id(node.id)

            graph_nodes.append((
                node_id,
                dict(
                    id = node_id,
                    chunk_id = str(chunk.chunk_id),
                    document_id = str(chunk.document_id),
                    name = node.name,
                    type = node.type.lower().capitalize(),
                    description = node.description,
                    created_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                    updated_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                )
            ))

            graph_edges.append((
                str(chunk.chunk_id),
                node_id,
                "contains",
                dict(
                    relationship_name = "contains",
                    source_node_id = str(chunk.chunk_id),
                    target_node_id = node_id,
                ),
            ))

            type_node_id = generate_node_id(node.type)

            if type_node_id not in existing_type_nodes_map:
                node_name = node.type.lower().capitalize()

                type_node = dict(
                    id = type_node_id,
                    name = node_name,
                    type = node_name,
                    created_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                    updated_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                )

                graph_nodes.append((type_node_id, type_node))
                existing_type_nodes_map[type_node_id] = type_node

            graph_edges.append((
                str(chunk.chunk_id),
                type_node_id,
                "contains_entity_type",
                dict(
                    relationship_name = "contains_entity_type",
                    source_node_id = str(chunk.chunk_id),
                    target_node_id = type_node_id,
                ),
            ))

            # Add relationship between entity type and entity itself: "Jake is Person"
            graph_edges.append((
                type_node_id,
                node_id,
                "is_entity_type",
                dict(
                    relationship_name = "is_entity_type",
                    source_node_id = type_node_id,
                    target_node_id = node_id,
                ),
            ))

            # Add relationship that came from graphs.
            for edge in graph.edges:
                graph_edges.append((
                    generate_node_id(edge.source_node_id),
                    generate_node_id(edge.target_node_id),
                    edge.relationship_name,
                    dict(
                        relationship_name = edge.relationship_name,
                        source_node_id = generate_node_id(edge.source_node_id),
                        target_node_id = generate_node_id(edge.target_node_id),
                    ),
                ))

    await graph_engine.add_nodes(graph_nodes)

    await graph_engine.add_edges(graph_edges)

    return data_chunks


def generate_node_id(node_id: str) -> str:
    return node_id.upper().replace(" ", "_").replace("'", "")






from pydantic import BaseModel, Field
from typing import Any, Dict, List, Optional, Union, Type

class Relationship(BaseModel):
    type: str = Field(..., description="The type of relationship, e.g., 'belongs_to'.")
    source: Optional[str] = Field(None, description="The identifier of the source id in the relationship.")
    target: Optional[str] = Field(None, description="The identifier of the target id in the relationship.")
    properties: Optional[Dict[str, Any]] = Field(None, description="A dictionary of additional properties related to the relationship.")

class Document(BaseModel):
    node_id: str
    title: str
    description: Optional[str] = None
    default_relationship: Relationship

class DirectoryModel(BaseModel):
    node_id: str
    path: str
    summary: str
    documents: List[Document] = []
    subdirectories: List['DirectoryModel'] = []
    default_relationship: Relationship

DirectoryModel.update_forward_refs()

class DirMetadata(BaseModel):
    node_id: str
    summary: str
    owner: str
    description: Optional[str] = None
    directories: List[DirectoryModel] = []
    documents: List[Document] = []
    default_relationship: Relationship

class GitHubRepositoryModel(BaseModel):
    node_id: str
    metadata: DirMetadata
    root_directory: DirectoryModel



class RelationshipModel(BaseModel):
    type: str
    source: str
    target: str

class NodeModel(BaseModel):
    node_id: str
    name: str
    default_relationship: Optional[RelationshipModel] = None
    children: List[Union[Dict[str, Any], "NodeModel"]] = Field(default_factory=list)
NodeModel.update_forward_refs()


""" This module contains the TopologyEngine class which is responsible for adding graph topology from a JSON or CSV file. """

import csv
import json
import logging
import os
from typing import Any, Dict, List, Optional, Union, Type

import asyncio
import aiofiles
import pandas as pd
from pydantic import BaseModel

from cognee.infrastructure.data.chunking.config import get_chunk_config
from cognee.infrastructure.data.chunking.get_chunking_engine import get_chunk_engine
from cognee.infrastructure.databases.graph.get_graph_engine import get_graph_engine
from cognee.infrastructure.databases.relational import get_relationaldb_config
from cognee.infrastructure.files.utils.extract_text_from_file import extract_text_from_file
from cognee.infrastructure.files.utils.guess_file_type import guess_file_type, FileTypeException
from cognee.modules.cognify.config import get_cognify_config
from cognee.base_config import get_base_config
from cognee.modules.topology.topology_data_models import NodeModel

cognify_config = get_cognify_config()
base_config = get_base_config()

logger = logging.getLogger("topology")

class TopologyEngine:
    def __init__(self, infer:bool) -> None:
        self.models: Dict[str, Type[BaseModel]] = {}
        self.infer = infer

    async def flatten_model(self, model: NodeModel, parent_id: Optional[str] = None) -> Dict[str, Any]:
        """Flatten the model to a dictionary."""
        result = model.dict()
        result["parent_id"] = parent_id
        if model.default_relationship:
            result.update({
                "relationship_type": model.default_relationship.type,
                "relationship_source": model.default_relationship.source,
                "relationship_target": model.default_relationship.target
            })
        return result

    async def recursive_flatten(self, items: Union[List[Dict[str, Any]], Dict[str, Any]], parent_id: Optional[str] = None) -> List[Dict[str, Any]]:
        """Recursively flatten the items.  """
        flat_list = []

        if isinstance(items, list):
            for item in items:
                flat_list.extend(await self.recursive_flatten(item, parent_id))
        elif isinstance(items, dict):
            model = NodeModel.parse_obj(items)
            flat_list.append(await self.flatten_model(model, parent_id))
            for child in model.children:
                flat_list.extend(await self.recursive_flatten(child, model.node_id))
        return flat_list

    async def load_data(self, file_path: str) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """Load data from a JSON or CSV file."""
        try:
            if file_path.endswith(".json"):
                async with aiofiles.open(file_path, mode="r") as f:
                    data = await f.read()
                    return json.loads(data)
            elif file_path.endswith(".csv"):
                async with aiofiles.open(file_path, mode="r") as f:
                    content = await f.read()
                    reader = csv.DictReader(content.splitlines())
                    return list(reader)
            else:
                raise ValueError("Unsupported file format")
        except Exception as e:
            raise RuntimeError(f"Failed to load data from {file_path}: {e}")

    async def add_graph_topology(self, file_path: str = None, files: list = None):
        """Add graph topology from a JSON or CSV file."""
        if self.infer:
            from cognee.modules.topology.infer_data_topology import infer_data_topology

            initial_chunks_and_ids = []

            chunk_config = get_chunk_config()
            chunk_engine = get_chunk_engine()
            chunk_strategy = chunk_config.chunk_strategy

            for base_file in files:
                with open(base_file["file_path"], "rb") as file:
                    try:
                        file_type = guess_file_type(file)
                        text = extract_text_from_file(file, file_type)

                        subchunks, chunks_with_ids = chunk_engine.chunk_data(chunk_strategy, text, chunk_config.chunk_size,
                                                                chunk_config.chunk_overlap)

                        if chunks_with_ids[0][0] == 1:
                            initial_chunks_and_ids.append({base_file["id"]: chunks_with_ids})

                    except FileTypeException:
                        logger.warning("File (%s) has an unknown file type. We are skipping it.", file["id"])


            topology = await infer_data_topology(str(initial_chunks_and_ids))
            graph_client = await get_graph_engine()

            await graph_client.add_nodes([(node["id"], node) for node in topology["nodes"]])
            await graph_client.add_edges((
                edge["source_node_id"],
                edge["target_node_id"],
                edge["relationship_name"],
                dict(relationship_name = edge["relationship_name"]),
            ) for edge in topology["edges"])

        else:
            dataset_level_information = files[0][1]

            # Extract the list of valid IDs from the explanations
            valid_ids = {item["id"] for item in dataset_level_information}
            try:
                data = await self.load_data(file_path)
                flt_topology = await self.recursive_flatten(data)
                df = pd.DataFrame(flt_topology)
                graph_client = await get_graph_engine()

                for _, row in df.iterrows():
                    node_data = row.to_dict()
                    node_id = node_data.pop("node_id", None)
                    if node_id in valid_ids:
                        await graph_client.add_node(node_id, node_data)
                    if node_id not in valid_ids:
                        raise ValueError(f"Node ID {node_id} not found in the dataset")
                    if pd.notna(row.get("relationship_source")) and pd.notna(row.get("relationship_target")):
                        await graph_client.add_edge(row["relationship_source"], row["relationship_target"], relationship_name=row["relationship_type"])

                return
            except Exception as e:
                raise RuntimeError(f"Failed to add graph topology from {file_path}: {e}") from e



async def main():
    # text = """Conservative PP in the lead in Spain, according to estimate
    #         An estimate has been published for Spain:
    #
    #         Opposition leader Alberto Núñez Feijóo’s conservative People’s party (PP): 32.4%
    #
    #         Spanish prime minister Pedro Sánchez’s Socialist party (PSOE): 30.2%
    #
    #         The far-right Vox party: 10.4%
    #
    #         In Spain, the right has sought to turn the European election into a referendum on Sánchez.
    #
    #         Ahead of the vote, public attention has focused on a saga embroiling the prime minister’s wife, Begoña Gómez, who is being investigated over allegations of corruption and influence-peddling, which Sanchez has dismissed as politically-motivated and totally baseless."""
    # text_two = """The far-right Vox party: 10.4%"""

    from cognee.api.v1.add import add
    dataset_name = "explanations"
    print(os.getcwd())
    data_dir = os.path.abspath("../../.data")
    print(os.getcwd())

    await add(f"data://{data_dir}", dataset_name="explanations")

    relational_config = get_relationaldb_config()
    db_engine = relational_config.database_engine


    datasets = db_engine.get_datasets()
    dataset_files =[]

    for added_dataset in datasets:
        if dataset_name in added_dataset:
            dataset_files.append((added_dataset, db_engine.get_files_metadata(added_dataset)))



    print(dataset_files)
    topology_engine = TopologyEngine(infer=True)
    file_path = "example_data.json"  # or 'example_data.csv'
    #
    # # Adding graph topology
    graph = await topology_engine.add_graph_topology(file_path, dataset_files=dataset_files)
    print(graph)

# Run the main function
if __name__ == "__main__":
    asyncio.run(main())
